from IPython.display import Math, display


class Symbol:

    def __init__(self, latex_string, tag=None):
        if isinstance(latex_string, Symbol):
            latex_string = str(latex_string)
        self.latex_string = latex_string
        if tag:
            assert isinstance(tag, str)
            _expr_store[tag] = self

    def __str__(self):
        return self.latex_string

    def __repr__(self):
        return str(self)

    def __call__(self, *args, **kwargs):
        s = self.latex_string + '\\left( '
        for i, arg in enumerate(args):
            s += str(arg)
            if i < len(args) - 1:
                s += ', '
        s += ' \\right)'
        return Symbol(s)

    def __pow__(self, power):
        s = '%s^{%s}' % (str(self), power)
        return Symbol(s)

    def __add__(self, other):
        s = '%s + %s' % (str(self), str(other))
        return Symbol(s)

    def __radd__(self, other):
        s = '%s + %s' % (str(other), str(self))
        return Symbol(s)

    def __eq__(self, other):
        s = '%s = %s' % (str(self), str(other))
        return Symbol(s)

    def __neg__(self):
        return Symbol('- %s' % self.latex_string)

    def __mul__(self, other):
        return Symbol('%s %s' % (str(self), str(other)))

    def __rmul__(self, other):
        return Symbol('%s %s' % (str(other), str(self)))

    def __sub__(self, other):
        s = '%s - %s' % (str(self), str(other))
        return Symbol(s)

    def __rsub__(self, other):
        s = '%s - %s' % (str(other), str(self))
        return Symbol(s)

    def __rshift__(self, other):
        assert isinstance(other, str)
        return self

    def _ipython_display_(self, **kwargs):
        display(Math(str(self)))

    def _repr_latex_(self):
        return Math(str(self))


class Int(Symbol):

    def __init__(self, lower='', upper='', **kwargs):
        super().__init__('\\int_{%s}^{%s}' % (str(lower), str(upper)), **kwargs)


class Frac(Symbol):

    def __init__(self, numer, denom, **kwargs):
        self.numer = Symbol(numer)
        self.denom = Symbol(denom)
        super().__init__('\\frac{%s}{%s}' % (self.numer, self.denom), **kwargs)


class Sqrt(Symbol):

    def __init__(self, arg, **kwargs):
        self.arg = Symbol(arg)
        super().__init__('\\sqrt{%s}' % self.arg, **kwargs)


class Matrix(Symbol):

    def __init__(self, nrows, ncols, default='', **kwargs):
        self.values = {}
        self.nrows = nrows
        self.ncols = ncols

        for i in range(nrows):
            for j in range(ncols):
                self.values[(i, j)] = Symbol(str(default))

        super().__init__(str(self), **kwargs)

    def __setitem__(self, idx, value):
        assert len(idx) == 2

        row_inds = self._expand_slice(idx[0], self.nrows)
        col_inds = self._expand_slice(idx[1], self.ncols)

        k = 0
        for i in row_inds:

            for j in col_inds:
                if hasattr(value, '__len__'):
                    self.values[i, j] = Symbol(str(value[k]))
                else:
                    self.values[i, j] = Symbol(str(value))
                k += 1

    def _expand_slice(self, slice_or_ind, max_vals):
        if isinstance(slice_or_ind, slice):
            if slice_or_ind.start is None:
                start = 0
            else:
                start = slice_or_ind.start
            if slice_or_ind.stop is None:
                stop = max_vals
            else:
                stop = slice_or_ind.stop
            inds = list(range(start, stop))
        else:
            inds = [slice_or_ind]
        return inds

    def __str__(self):
        s = r'\left(\begin{matrix}'
        for i in range(self.nrows):
            for j in range(self.ncols):
                s += str(self.values[i, j])
                if j < self.ncols - 1:
                    s += ' & '
                elif i < self.nrows - 1:
                    s += r'\\'
        s += r'\end{matrix}\right)'
        return s

    def __repr__(self):
        return self.__str__()


def symbols(string):
    parts = string.split(',')
    if len(parts) == 1:
        return Symbol(string)
    return [Symbol(part.strip()) for part in parts]


def brace(expr):
    return r'\left(' * expr * r'\right)'

def get(tag):
    return _expr_store.get(tag)


_expr_store = {}

oo = Symbol(r'\infty')


from ..symbols import Symbol

_letters = (a, b, c, d, e, f,
            g, h, i, j, k, l,
            m, n, o, p, q, r,
            s, t, u, v, w, x,
            y, z) = tuple(Symbol(s) for s in 'abcdefghijklmnopqrstuvwxyz')

_differentials = (da, db, dc, dd, de, df,
                  dg, dh, di, dj, dk, dl,
                  dm, dn, do, dp, dq, dr,
                  ds, dt, du, dv, dw, dx,
                  dy, dz) = tuple(Symbol('d' + s) for s in 'abcdefghijklmnopqrstuvwxyz')


import unittest

from swiftex import *
from swiftex.abc import *


class TestSymbols(unittest.TestCase):

    def test_create_single_symbol_assigns_string(self):
        a = symbols('a')
        expected = 'a'
        actual = str(a)
        self.assertEqual(expected, actual)

    def test_create_multiple_symbols(self):
        b, c = symbols('b, c')
        self.assertEqual('b', str(b))
        self.assertEqual('c', str(c))

    def test_symbol_raised_to_power(self):
        # Arrange
        a = symbols('a')

        # Act
        actual = a ** 2

        # Assert
        self.assertEqual('a^{2}', str(actual))

    def test_adding_symbols(self):
        # Arrange
        a, b = symbols('a, b')

        # Act
        actual = a ** 2 + b ** 2

        # Assert
        self.assertEqual('a^{2} + b^{2}', str(actual))

    def test_pythagoras(self):
        # Arrange
        a, b, c = symbols('a, b, c')

        # Act
        actual = a ** 2 + b ** 2 == c ** 2

        # Assert
        self.assertEqual('a^{2} + b^{2} = c^{2}', str(actual))

    def test_abc(self):
        # Act
        actual = x ** 2 + y ** 2 == z ** 2

        # Assert
        self.assertEqual('x^{2} + y^{2} = z^{2}', str(actual))

    def test_attach_symbols(self):

        actual = a * b * c
        self.assertEqual('a b c', str(actual))


class TestIntegral(unittest.TestCase):

    def test_integral_sign_with_limits(self):
        expr = Int(-oo, oo)
        self.assertEqual(r'\int_{-\infty}^{\infty}', expr)

    def test_integral_x_squared_dx(self):
        expr = Int(a, b) * x**2 * dx
        self.assertEqual(r'\int_{a}^{b} x^{2} dx', expr)

    def test_integral_f_of_x(self):
        expr = Int(a, b) * f(x) * dx
        self.assertEqual(r'\int_{a}^{b} f \left( x \right) ^{2} dx', expr)

    def test_integral_f_of_x_and_y(self):
        expr = Int(a, b) * f(x, y) * dx
        self.assertEqual(r'\int_{a}^{b} f \left( x, y \right) ^{2} dx', expr)

    def test_integral_only_upper_limit(self):
        expr = Int('', b) * f(x, y) * dx
        self.assertEqual(r'\int_{}^{b} f \left( x, y \right) ^{2} dx', expr)

    def test_integral_only_lower_limit(self):
        expr = Int(a, '') * f(x, y) * dx
        self.assertEqual(r'\int_{}^{} f \left( x, y \right) ^{2} dx', expr)

    def test_integral_no_limits(self):
        expr = Int() * f(x, y) * dx
        self.assertEqual(r'\int f \left( x, y \right) ^{2} dx', expr)


class TestFrac(unittest.TestCase):

    def test_simple_fraction_as_strings(self):
        actual = Frac('a + b', 'c + d')
        self.assertEqual(r'\frac{a + b}{c + d}', actual)

    def test_simple_fraction_as_symbols(self):
        actual = Frac(a+b, c+d)
        self.assertEqual(r'\frac{a + b}{c + d}', actual)


class TestSqrt(unittest.TestCase):

    def test_sqrt_with_str_arg(self):
        actual = Sqrt('b^2 - 4ac')
        self.assertEqual(r'\sqrt{b^2 - 4ac}', actual)

    def test_sqrt_with_symbol_arg(self):
        actual = Sqrt(b**2 - 4 * a * c)
        self.assertEqual(r'\sqrt{b^2 - 4 a c}', actual)


class TestTagging(unittest.TestCase):

    def test_tag_and_retrieve_subexpression(self):
        Frac(-b + Sqrt(b**2 - 4*a*c, tag='the_root'), 2*a)
        actual = get('the_root')
        self.assertEqual(str(Sqrt(b**2 - 4*a*c)), actual)


class TestMatrix(unittest.TestCase):

    def test_constant_matrix(self):
        A = Matrix(2, 2, default=1)
        expected = r'\left(' \
                   r'\begin{matrix}' \
                   r'1 & 1 \\' \
                   r'1 & 1 ' \
                   r'\end{matrix}' \
                   r'\right)'
        self.assertEqual(expected, A)

    def test_set_matrix_by_index(self):
        A = Matrix(2, 2, default=0)
        A[0, 1] = 1
        A[1, 0] = 1
        expected = r'\left(' \
                   r'\begin{matrix}' \
                   r'0 & 1 \\' \
                   r'1 & 0 ' \
                   r'\end{matrix}' \
                   r'\right)'
        self.assertEqual(expected, A)

    def test_matrix_slice_with_constant(self):
        A = Matrix(3, 3, default=0)
        A[1, :] = 1
        expected = B = Matrix(3, 3, default=0)
        B[1, 0] = 1
        B[1, 1] = 1
        B[1, 2] = 1
        self.assertEqual(str(expected), str(A))

    def test_matrix_slice_with_sequence(self):
        A = Matrix(3, 3, default=0)
        A[1, :] = 1, 2, 3
        expected = B = Matrix(3, 3, default=0)
        B[1, 0] = 1
        B[1, 1] = 2
        B[1, 2] = 3
        self.assertEqual(str(expected), str(A))

    def test_brace(self):
        actual = brace(a+b)**2
        self.assertEqual('(a + b)^2', actual)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
from os.path import exists
from setuptools import setup, find_packages

author = 'mathcube'
email = 'mathcube7@gmail.com'
description = 'Easy and fast math typesetting with Python'
name = 'swiftex'
year = '2021'
url = 'https://github.com/mathcube7/swiftex'
version = '0.0.4'

setup(
    name=name,
    author=author,
    author_email=email,
    url=url,
    version=version,
    packages=find_packages(),
    package_dir={name: name},
    include_package_data=True,
    license='MIT',
    description=description,
    long_description=open('README.md').read() if exists('README.md') else '',
    long_description_content_type="text/markdown",
    install_requires=['sphinx', 'IPython',
                      ],
    python_requires=">=3.6",
    classifiers=['Operating System :: OS Independent',
                 'Programming Language :: Python :: 3',
                 ],
    platforms=['ALL'],
)
